from __future__ import annotations

import re
from functools import reduce
from typing import TYPE_CHECKING, Literal, NamedTuple

from django.db.models import (
    IntegerField,
    Manager,
    Q,
    QuerySet,
    TextField,
    Value,
)
from django.db.models.functions import Cast, Coalesce
from django.db.models.lookups import (
    Contains,
    Exact,
    IContains,
    IExact,
    IRegex,
    IStartsWith,
    Regex,
    StartsWith,
)
from lamin_utils import logger
from lamin_utils._lookup import Lookup
from lamindb_setup.core import deprecated
from lamindb_setup.core._docs import doc_args

if TYPE_CHECKING:
    from ..base.types import StrField


def _search(
    cls,
    string: str,
    *,
    field: StrField | list[StrField] | None = None,
    limit: int | None = 20,
    case_sensitive: bool = False,
    truncate_string: bool = False,
) -> QuerySet:
    """Search.

    Args:
        string: The input string to match against the field ontology values.
        field: The field or fields to search. Search all string fields by default.
        limit: Maximum amount of top results to return.
        case_sensitive: Whether the match is case sensitive.

    Returns:
        A sorted `DataFrame` of search results with a score in column `score`.
        If `return_queryset` is `True`.  `QuerySet`.

    See Also:
        :meth:`~lamindb.models.SQLRecord.filter`
        :meth:`~lamindb.models.SQLRecord.lookup`

    Examples:
        >>> ulabels = ln.ULabel.from_values(["ULabel1", "ULabel2", "ULabel3"], field="name")
        >>> ln.save(ulabels)
        >>> ln.ULabel.search("ULabel2")
    """
    if string is None:
        raise ValueError("Cannot search for None value! Please pass a valid string.")

    input_queryset = (
        cls.all() if isinstance(cls, (QuerySet, Manager)) else cls.objects.all()
    )
    registry = input_queryset.model
    name_field = getattr(registry, "_name_field", "name")
    if field is None:
        fields = [
            field.name
            for field in registry._meta.fields
            if field.get_internal_type() in {"CharField", "TextField"}
        ]
    else:
        if not isinstance(field, list):
            fields_input = [field]
        else:
            fields_input = field
        fields = []
        for field in fields_input:
            if not isinstance(field, str):
                try:
                    fields.append(field.field.name)
                except AttributeError as error:
                    raise TypeError(
                        "Please pass a SQLRecord string field, e.g., `CellType.name`!"
                    ) from error
            else:
                fields.append(field)

    if truncate_string:
        if (len_string := len(string)) > 5:
            n_80_pct = int(len_string * 0.8)
            string = string[:n_80_pct]

    string = string.strip()
    string_escape = re.escape(string)

    exact_lookup = Exact if case_sensitive else IExact
    regex_lookup = Regex if case_sensitive else IRegex
    contains_lookup = Contains if case_sensitive else IContains

    ranks = []
    contains_filters = []
    for field in fields:
        field_expr = Coalesce(
            Cast(field, output_field=TextField()),
            Value(""),
            output_field=TextField(),
        )
        # exact rank
        exact_expr = exact_lookup(field_expr, string)
        exact_rank = Cast(exact_expr, output_field=IntegerField()) * 200
        ranks.append(exact_rank)
        # exact synonym
        synonym_expr = regex_lookup(field_expr, rf"(?:^|.*\|){string_escape}(?:\|.*|$)")
        synonym_rank = Cast(synonym_expr, output_field=IntegerField()) * 200
        ranks.append(synonym_rank)
        # match as sub-phrase
        sub_expr = regex_lookup(
            field_expr, rf"(?:^|.*[ \|\.,;:]){string_escape}(?:[ \|\.,;:].*|$)"
        )
        sub_rank = Cast(sub_expr, output_field=IntegerField()) * 10
        ranks.append(sub_rank)
        # startswith and avoid matching string with " " on the right
        # mostly for truncated
        startswith_expr = regex_lookup(
            field_expr, rf"(?:^|.*\|){string_escape}[^ ]*(?:\|.*|$)"
        )
        startswith_rank = Cast(startswith_expr, output_field=IntegerField()) * 8
        ranks.append(startswith_rank)
        # match as sub-phrase from the left, mostly for truncated
        right_expr = regex_lookup(field_expr, rf"(?:^|.*[ \|]){string_escape}.*")
        right_rank = Cast(right_expr, output_field=IntegerField()) * 2
        ranks.append(right_rank)
        # match as sub-phrase from the right
        left_expr = regex_lookup(field_expr, rf".*{string_escape}(?:$|[ \|\.,;:].*)")
        left_rank = Cast(left_expr, output_field=IntegerField()) * 2
        ranks.append(left_rank)
        # simple contains filter
        contains_expr = contains_lookup(field_expr, string)
        contains_filter = Q(contains_expr)
        contains_filters.append(contains_filter)
        # also rank by contains
        contains_rank = Cast(contains_expr, output_field=IntegerField())
        ranks.append(contains_rank)
        # additional rule for truncated strings
        # weight matches from the beginning of the string higher
        # sometimes whole words get truncated and startswith_expr is not enough
        if truncate_string and field == name_field:
            startswith_lookup = StartsWith if case_sensitive else IStartsWith
            name_startswith_expr = startswith_lookup(field_expr, string)
            name_startswith_rank = (
                Cast(name_startswith_expr, output_field=IntegerField()) * 2
            )
            ranks.append(name_startswith_rank)

    ranked_queryset = (
        input_queryset.filter(reduce(lambda a, b: a | b, contains_filters))
        .alias(rank=sum(ranks))
        .order_by("-rank")
    )

    return ranked_queryset[:limit]


def _lookup(
    cls,
    field: StrField | None = None,
    return_field: StrField | None = None,
    using_key: str | None = None,
    keep: Literal["first", "last", False] = "first",
) -> NamedTuple:
    """Return an auto-complete object for a field.

    Args:
        field: The field to look up the values for. Defaults to first string field.
        return_field: The field to return. If `None`, returns the whole record.
        keep: When multiple records are found for a lookup, how to return the records.
            - `"first"`: return the first record.
            - `"last"`: return the last record.
            - `False`: return all records.

    Returns:
        A `NamedTuple` of lookup information of the field values with a
        dictionary converter.

    See Also:
        :meth:`~lamindb.models.SQLRecord.search`

    Examples:
        >>> import bionty as bt
        >>> bt.settings.organism = "human"
        >>> bt.Gene.from_source(symbol="ADGB-DT").save()
        >>> lookup = bt.Gene.lookup()
        >>> lookup.adgb_dt
        >>> lookup_dict = lookup.dict()
        >>> lookup_dict['ADGB-DT']
        >>> lookup_by_ensembl_id = bt.Gene.lookup(field="ensembl_gene_id")
        >>> genes.ensg00000002745
        >>> lookup_return_symbols = bt.Gene.lookup(field="ensembl_gene_id", return_field="symbol")
    """
    from .sqlrecord import get_name_field

    queryset = cls.all() if isinstance(cls, (QuerySet, Manager)) else cls.objects.all()
    field = get_name_field(registry=queryset.model, field=field)

    return Lookup(
        records=queryset,
        values=[i.get(field) for i in queryset.values()],
        tuple_name=cls.__class__.__name__,
        prefix="ln",
        keep=keep,
    ).lookup(
        return_field=(
            get_name_field(registry=queryset.model, field=return_field)
            if return_field is not None
            else None
        )
    )


# this is the default (._default_manager and ._base_manager) for lamindb models
class QueryManager(Manager):
    """Manage queries through fields.

    See Also:

        :class:`lamindb.models.QuerySet`
        `django Manager <https://docs.djangoproject.com/en/4.2/topics/db/managers/>`__

    Examples:

        >>> ln.save(ln.ULabel.from_values(["ULabel1", "ULabel2", "ULabel3"], field="name"))  # noqa
        >>> labels = ln.ULabel.filter(name__icontains = "label")
        >>> ln.ULabel(name="ULabel1").save()
        >>> label = ln.ULabel.get(name="ULabel1")
        >>> label.parents.set(labels)
        >>> manager = label.parents
        >>> manager.to_dataframe()
    """

    def track_run_input_manager(self):
        if hasattr(self, "source_field_name") and hasattr(self, "target_field_name"):
            if (
                self.source_field_name == "collection"
                and self.target_field_name == "artifact"
            ):
                from lamindb import settings
                from lamindb.core._context import context
                from lamindb.models.artifact import (
                    WARNING_RUN_TRANSFORM,
                    track_run_input,
                )

                if (
                    context.run is None
                    and not settings.creation.artifact_silence_missing_run_warning
                ):
                    logger.warning(WARNING_RUN_TRANSFORM)
                track_run_input(self.instance)

    def to_list(self, field: str | None = None):
        """Populate a list with the results.

        Examples:
            >>> ln.save(ln.ULabel.from_values(["ULabel1", "ULabel2", "ULabel3"], field="name"))
            >>> labels = ln.ULabel.filter(name__icontains="label")
            >>> ln.ULabel(name="ULabel1").save()
            >>> label = ln.ULabel.get(name="ULabel1")
            >>> label.parents.set(labels)
            >>> label.parents.to_list()
            >>> label.parents.to_list("name")
            ['ULabel1', 'ULabel2', 'ULabel3']
        """
        if field is None:
            return list(self.all())
        else:
            self.track_run_input_manager()
            return list(self.values_list(field, flat=True))

    @deprecated(new_name="to_list")
    def list(self, field: str | None = None):
        return self.to_list(field)

    def to_dataframe(self, **kwargs):
        """Convert to DataFrame.

        For `**kwargs`, see :meth:`lamindb.models.QuerySet.to_dataframe`.
        """
        return self.all().to_dataframe(**kwargs)

    @deprecated(new_name="to_dataframe")
    def df(self, **kwargs):
        return self.to_dataframe(**kwargs)

    def all(self):
        """Return QuerySet of all.

        For `**kwargs`, see :meth:`lamindb.models.QuerySet.to_dataframe`.
        """
        self.track_run_input_manager()
        return super().all()

    @doc_args(_search.__doc__)
    def search(self, string: str, **kwargs):
        """{}"""  # noqa: D415
        return _search(cls=self.all(), string=string, **kwargs)

    @doc_args(_lookup.__doc__)
    def lookup(self, field: StrField | None = None, **kwargs) -> NamedTuple:
        """{}"""  # noqa: D415
        return _lookup(cls=self.all(), field=field, **kwargs)

    def get_queryset(self):
        from .query_set import BasicQuerySet

        # QueryManager returns BasicQuerySet because it is problematic to redefine .filter and .get
        # for a query set used by the default manager
        return BasicQuerySet(model=self.model, using=self._db, hints=self._hints)



# Generated by Django 5.1.12 on 2025-09-28 21:39

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0122_remove_personproject_person_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifact",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="branch",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="collection",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="feature",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="record",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="schema",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="space",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="storage",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="transform",
            name="description",
            field=lamindb.base.fields.TextField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="key",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=1024, null=True
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="key",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=1024, null=True
            ),
        ),
    ]



# Generated by Django 5.2 on 2025-06-02 10:53

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields
import lamindb.base.uids


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0103_remove_writelog_migration_state_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="branch",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.uids.base62_12,
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="recordrecord",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="recordrecord",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="sheet",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="sheet",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="branch",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
    ]



from __future__ import annotations

from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Any, TypeAlias

from anndata import AnnData
from pandas import DataFrame

from lamindb.core._compat import (
    with_package_obj,
)
from lamindb.core.types import ScverseDataStructures

if TYPE_CHECKING:
    from lamindb_setup.types import UPathStr

SupportedDataTypes: TypeAlias = DataFrame | ScverseDataStructures


def infer_suffix(dmem: SupportedDataTypes, format: str | dict[str, Any] | None = None):
    """Infer LaminDB storage file suffix from a data object."""
    if isinstance(dmem, AnnData):
        assert not isinstance(format, dict)  # noqa: S101
        if format is not None:
            # should be `.h5ad`, `.`zarr`, or `.anndata.zarr`
            if format not in {"h5ad", "zarr", "anndata.zarr"}:
                raise ValueError(
                    "Error when specifying AnnData storage format, it should be"
                    f" 'h5ad', 'zarr', not '{format}'. Check 'format'"
                    " or the suffix of 'key'."
                )
            return "." + format
        return ".h5ad"

    if isinstance(dmem, DataFrame):
        if isinstance(format, str):
            if format == ".csv":
                return ".csv"
        elif isinstance(format, dict):
            if format.get("suffix") == ".csv":
                return ".csv"
        return ".parquet"

    if with_package_obj(
        dmem,
        "MuData",
        "mudata",
        lambda obj: True,  # Just checking type, not calling any method
    )[0]:
        return ".h5mu"

    has_spatialdata, spatialdata_suffix = with_package_obj(
        dmem,
        "SpatialData",
        "spatialdata",
        lambda obj: (
            format
            if format is not None and format in {"spatialdata.zarr", "zarr"}
            else ".zarr"
            if format is None
            else (_ for _ in ()).throw(
                ValueError(
                    "Error when specifying SpatialData storage format, it should be"
                    f" 'zarr', 'spatialdata.zarr', not '{format}'. Check 'format'"
                    " or the suffix of 'key'."
                )
            )
        ),
    )
    if has_spatialdata:
        return spatialdata_suffix
    else:
        raise NotImplementedError


def write_to_disk(dmem: SupportedDataTypes, filepath: UPathStr, **kwargs) -> None:
    """Writes the passed in memory data to disk to a specified path."""
    if isinstance(dmem, AnnData):
        suffix = PurePosixPath(filepath).suffix
        if suffix == ".h5ad":
            dmem.write_h5ad(filepath)
            return
        elif suffix == ".zarr":
            dmem.write_zarr(filepath)
            return
        else:
            raise NotImplementedError

    if isinstance(dmem, DataFrame):
        if filepath.suffix == ".csv":
            dmem.to_csv(filepath, **kwargs)
            return
        dmem.to_parquet(filepath, **kwargs)
        return

    if with_package_obj(dmem, "MuData", "mudata", lambda obj: obj.write(filepath))[0]:
        return

    if with_package_obj(
        dmem,
        "SpatialData",
        "spatialdata",
        lambda obj: obj.write(filepath, overwrite=True),
    )[0]:
        return

    raise NotImplementedError



"""Examples for MLCommons Croissant files, which are used to store metadata about datasets.

.. autosummary::
   :toctree: .

   mini_immuno

"""

import json
from pathlib import Path


def mini_immuno(
    n_files: int = 1, filepath_prefix: str = "", strip_version: bool = False
) -> list[Path]:
    """Return paths to the mini immuno dataset and its metadata as a Croissant file.

    Args:
        n_files: Number of files inside the croissant file.
        filepath_prefix: Move the dataset and references to it in a specific directory.

    Example

        ::

            croissant_path, dataset1_path = ln.examples.croissant.mini_immuno()
            croissant_path, dataset1_path, dataset2_path = ln.examples.croissant.mini_immuno(n_files=2)
    """
    from ..datasets import file_mini_csv
    from ..datasets.mini_immuno import get_dataset1

    adata = get_dataset1(otype="AnnData")
    if filepath_prefix:
        dataset1_path = Path(filepath_prefix) / "mini_immuno.anndata.zarr"
    else:
        dataset1_path = Path("mini_immuno.anndata.zarr")
    adata.write_zarr(dataset1_path)
    orig_croissant_path = (
        Path(__file__).parent / "mini_immuno.anndata.zarr_metadata.json"
    )
    with open(orig_croissant_path, encoding="utf-8") as f:
        data = json.load(f)
    if filepath_prefix:
        assert data["distribution"][0]["@id"] == "mini_immuno.anndata.zarr"  # noqa: S101
        data["distribution"][0]["@id"] = str(Path(filepath_prefix) / dataset1_path.name)
    if strip_version:
        data.pop("version", None)
    if n_files == 2:
        file_mini_csv()
        if filepath_prefix:
            dataset2_path = Path(filepath_prefix) / "mini.csv"
        else:
            dataset2_path = Path("mini.csv")
        data["distribution"].append(
            {
                "@type": "sc:FileObject",
                "@id": dataset2_path.as_posix(),
                "name": "mini.csv",
                "encodingFormat": "text/csv",
            }
        )
    croissant_path = Path("mini_immuno.anndata.zarr_metadata.json")
    with open(croissant_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)

    result: list[Path] = [croissant_path, dataset1_path]
    if n_files == 1:
        return result
    result.append(dataset2_path)

    return result



# Generated by Django 5.2 on 2025-05-26 09:33

import django.db.models.deletion
from django.db import migrations, models
from lamindb_setup.core.hashing import hash_dict, hash_string

import lamindb.base.fields


def populate_hash_values(apps, schema_editor):
    from tqdm import tqdm

    FeatureValue = apps.get_model("lamindb", "FeatureValue")

    records_queryset = FeatureValue.objects.filter(hash__isnull=True)
    total_count = records_queryset.count()
    print(f"\nFound {total_count} FeatureValue records with hash=None. Populating...")

    # Convert to list to avoid QuerySet modification issues
    records_to_update = list(records_queryset)

    batch_size = 1000
    updated_count = 0

    for batch_start in tqdm(range(0, len(records_to_update), batch_size)):
        batch_records = records_to_update[batch_start : batch_start + batch_size]

        # Calculate hashes for all records in batch
        for record in batch_records:
            if not isinstance(record.value, dict):
                record.hash = hash_string(str(record.value))
            else:
                record.hash = hash_dict(record.value)

        # Bulk update the entire batch
        FeatureValue.objects.bulk_update(batch_records, ["hash"], batch_size=batch_size)
        updated_count += len(batch_records)

    print(f"Finished. Updated: {updated_count} records.")


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0100_branch_alter_artifact__branch_code_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifact",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="recordrecord",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="sheet",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="_branch_code",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_column="_branch_code",
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to="lamindb.branch",
            ),
        ),
        migrations.RenameField(
            model_name="artifact",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="collection",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="feature",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="featurevalue",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="person",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="project",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="record",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="recordrecord",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="run",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="schema",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="sheet",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="storage",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="transform",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.RenameField(
            model_name="ulabel",
            old_name="_branch_code",
            new_name="branch",
        ),
        migrations.AlterField(
            model_name="artifact",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=22,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="name",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=150
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="features",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="projects",
                to="lamindb.project",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="records",
                to="lamindb.record",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="references",
                to="lamindb.reference",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="schemas",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="transform",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=22, null=True
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="ulabels",
                to="lamindb.ulabel",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="transform",
            unique_together={("key", "hash")},
        ),
        migrations.AddConstraint(
            model_name="artifact",
            constraint=models.UniqueConstraint(
                condition=models.Q(("key__isnull", False)),
                fields=("storage", "key", "hash"),
                name="unique_artifact_storage_key_hash",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="_expect_many",
            field=models.BooleanField(db_default=None, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=lamindb.base.uids.base62_16,
                editable=False,
                max_length=20,
                unique=True,
            ),
        ),
        migrations.RemoveConstraint(
            model_name="featurevalue",
            name="unique_simple_feature_value",
        ),
        migrations.RemoveConstraint(
            model_name="featurevalue",
            name="unique_complex_feature_value",
        ),
        migrations.AlterUniqueTogether(
            name="featurevalue",
            unique_together={("feature", "hash")},
        ),
        migrations.RunPython(populate_hash_values),
    ]



import lamindb as ln
import pytest

TEST_URL = "https://github.com/openproblems-bio/task_batch_integration"


def test_transform_from_git():
    # test auto-inferred latest commit hash
    transform1 = ln.Transform.from_git(url=TEST_URL, path="main.nf")
    assert transform1.source_code.startswith(f"""\
repo: {TEST_URL}
path: main.nf
commit:""")
    assert transform1.key == "openproblems-bio/task_batch_integration/main.nf"
    assert transform1.version is None
    assert transform1.reference.startswith(f"{TEST_URL}/blob/")
    assert transform1.reference_type == "url"

    # test checking out specific version
    transform2 = ln.Transform.from_git(url=TEST_URL, path="main.nf", version="v2.0.0")
    assert transform2.source_code.startswith(f"""\
repo: {TEST_URL}
path: main.nf
commit:""")
    assert transform2.version == "v2.0.0"
    assert transform1.source_code != transform2.source_code
    assert transform1.reference != transform2.reference

    # test sliding transform from branch
    transform3 = ln.Transform.from_git(
        url=TEST_URL, path="main.nf", version="main", branch="main"
    )
    assert transform3.source_code.startswith(f"""\
repo: {TEST_URL}
path: main.nf
branch:""")


def test_transform_from_git_with_entrypoint():
    # test auto-inferred latest commit hash
    transform1 = ln.Transform.from_git(
        url=TEST_URL, path="main.nf", entrypoint="myentrypoint"
    )
    assert transform1.source_code.startswith(f"""\
repo: {TEST_URL}
path: main.nf
entrypoint: myentrypoint
commit:""")


def test_transform_custom_key():
    # test auto-inferred latest commit hash
    transform1 = ln.Transform.from_git(url=TEST_URL, path="main.nf", key="mypipeline")
    assert transform1.key == "mypipeline"


def test_transform_from_git_failure_modes():
    # invalid tag
    with pytest.raises(ValueError) as error:
        ln.Transform.from_git(
            url=TEST_URL,
            path="main.nf",
            version="invalid",
        )
    assert error.exconly().startswith("ValueError: Failed to checkout version invalid")

    # invalid branch
    with pytest.raises(ValueError) as error:
        ln.Transform.from_git(
            url=TEST_URL,
            path="main.nf",
            branch="invalid",
        )
    assert error.exconly().startswith("ValueError: Failed to checkout branch invalid")



import lamindb as ln

study_metadata_schema = ln.Schema(
    name="Study metadata schema",
    features=[
        ln.Feature(name="temperature", dtype=float).save(),
        ln.Feature(name="experiment", dtype=str).save(),
    ],
).save()



# Generated by Django 5.2 on 2025-01-13 05:55

import django.core.validators
import django.db.models.deletion
import lamindb_setup
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.ids
import lamindb.base.users
import lamindb.models


def migrate_data(apps, schema_editor):
    """Check if source table exists and run migration if it does."""
    db = schema_editor.connection
    cursor = db.cursor()

    # Check if table exists - works in both SQLite and PostgreSQL
    if db.vendor == "sqlite":
        cursor.execute("""
            SELECT count(*)
            FROM sqlite_master
            WHERE type='table' AND name='ourprojects_reference';
        """)
    else:  # postgresql
        cursor.execute("""
            SELECT count(*)
            FROM information_schema.tables
            WHERE table_schema = 'public'
            AND table_name = 'ourprojects_reference';
        """)

    table_exists = cursor.fetchone()[0] > 0
    if not table_exists:
        return

    # Get initial counts
    cursor.execute("SELECT COUNT(*) FROM ourprojects_reference")
    old_count = cursor.fetchone()[0]
    cursor.execute("SELECT COUNT(*) FROM lamindb_reference")
    initial_target_count = cursor.fetchone()[0]

    # Begin transaction
    cursor.execute("BEGIN")
    try:
        # Copy core tables, adding empty JSON for aux field
        cursor.execute("""
            INSERT INTO lamindb_reference(
                id, created_at, updated_at, created_by_id, uid, name, abbr, url,
                pubmed_id, doi, preprint, public, journal, description, text,
                published_at
            )
            SELECT
                id, created_at, updated_at, created_by_id, uid, name, abbr, url,
                pubmed_id, doi, preprint, public, journal, description, text,
                published_at
            FROM ourprojects_reference
        """)

        cursor.execute("""
            INSERT INTO lamindb_person(
                id, created_at, updated_at, created_by_id, uid, name, email,
                external
            )
            SELECT
                id, created_at, updated_at, created_by_id, uid, name, email,
                external
            FROM ourprojects_person
        """)

        cursor.execute("""
            INSERT INTO lamindb_project(
                id, created_at, updated_at, created_by_id, uid, name, abbr,
                url
            )
            SELECT
                id, created_at, updated_at, created_by_id, uid, name, abbr,
                url
            FROM ourprojects_project
        """)

        # Copy many-to-many relationships with aux field
        cursor.execute("""
            INSERT INTO lamindb_artifactreference(
                id, created_at, created_by_id, artifact_id, reference_id,
                feature_id, label_ref_is_name, feature_ref_is_name
            )
            SELECT
                id, created_at, created_by_id, artifact_id, reference_id,
                feature_id, label_ref_is_name, feature_ref_is_name
            FROM ourprojects_artifactreference
        """)

        cursor.execute("""
            INSERT INTO lamindb_transformreference(
                id, created_at, created_by_id, transform_id, reference_id
            )
            SELECT
                id, created_at, created_by_id, transform_id, reference_id
            FROM ourprojects_transformreference
        """)

        cursor.execute("""
            INSERT INTO lamindb_collectionreference(
                id, created_at, created_by_id, collection_id, reference_id
            )
            SELECT
                id, created_at, created_by_id, collection_id, reference_id
            FROM ourprojects_collectionreference
        """)

        cursor.execute("""
            INSERT INTO lamindb_artifactproject(
                id, created_at, created_by_id, artifact_id, project_id,
                feature_id, label_ref_is_name, feature_ref_is_name
            )
            SELECT
                id, created_at, created_by_id, artifact_id, project_id,
                feature_id, label_ref_is_name, feature_ref_is_name
            FROM ourprojects_artifactproject
        """)

        cursor.execute("""
            INSERT INTO lamindb_transformproject(
                id, created_at, created_by_id, transform_id, project_id
            )
            SELECT
                id, created_at, created_by_id, transform_id, project_id
            FROM ourprojects_transformproject
        """)

        cursor.execute("""
            INSERT INTO lamindb_collectionproject(
                id, created_at, created_by_id, collection_id, project_id
            )
            SELECT
                id, created_at, created_by_id, collection_id, project_id
            FROM ourprojects_collectionproject
        """)

        # Verify migration
        cursor.execute("SELECT COUNT(*) FROM lamindb_reference")
        final_count = cursor.fetchone()[0]
        expected_count = initial_target_count + old_count

        if final_count == expected_count:
            # Clean up ourprojects content
            cursor.execute("DELETE FROM django_migrations WHERE app = 'ourprojects'")

            # Drop tables - using standard SQL
            tables = [
                "ourprojects_reference",
                "ourprojects_person",
                "ourprojects_project",
                "ourprojects_artifactreference",
                "ourprojects_transformreference",
                "ourprojects_collectionreference",
                "ourprojects_artifactproject",
                "ourprojects_transformproject",
                "ourprojects_collectionproject",
            ]

            for table in tables:
                if db.vendor == "sqlite":
                    cursor.execute(f"DROP TABLE IF EXISTS {table}")
                else:  # postgresql
                    cursor.execute(f"DROP TABLE IF EXISTS {table} CASCADE")

            cursor.execute("COMMIT")
            print(
                "Data migration from ourprojects to lamindb successful, you can now access ourprojects data through lamindb"
            )
        else:
            cursor.execute("ROLLBACK")
            raise Exception("Migration failed: SQLRecord count mismatch")

    except Exception as e:
        cursor.execute("ROLLBACK")
        raise e


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0072_remove_user__branch_code_remove_user_aux_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="ArtifactProject",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "label_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "feature_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_project",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifactproject",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.CreateModel(
            name="ArtifactReference",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "label_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "feature_ref_is_name",
                    lamindb.base.fields.BooleanField(
                        blank=True, default=None, null=True
                    ),
                ),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_reference",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifactreference",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.CreateModel(
            name="CollectionProject",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "collection",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_project",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.CreateModel(
            name="CollectionReference",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "collection",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_reference",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                ("aux", models.JSONField(db_default=None, default=None, null=True)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_8,
                        max_length=8,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=255
                    ),
                ),
                (
                    "email",
                    lamindb.base.fields.EmailField(
                        blank=True, default=None, max_length=254, null=True
                    ),
                ),
                (
                    "external",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                lamindb.models.CanCurate,
                models.Model,
                lamindb.models.ValidateFields,
            ),
        ),
        migrations.CreateModel(
            name="Project",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                ("aux", models.JSONField(db_default=None, default=None, null=True)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_12,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=255
                    ),
                ),
                (
                    "abbr",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "url",
                    lamindb.base.fields.URLField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="projects",
                        through="lamindb.ArtifactProject",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="projects",
                        through="lamindb.CollectionProject",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "contributors",
                    models.ManyToManyField(
                        related_name="projects", to="lamindb.person"
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                lamindb.models.CanCurate,
                models.Model,
                lamindb.models.ValidateFields,
            ),
        ),
        migrations.AddField(
            model_name="collectionproject",
            name="project",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_collection",
                to="lamindb.project",
            ),
        ),
        migrations.AddField(
            model_name="artifactproject",
            name="project",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="lamindb.project",
            ),
        ),
        migrations.CreateModel(
            name="Reference",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(auto_now=True, db_index=True),
                ),
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                ("aux", models.JSONField(db_default=None, default=None, null=True)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_12,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=255
                    ),
                ),
                (
                    "abbr",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("url", lamindb.base.fields.URLField(blank=True, null=True)),
                (
                    "pubmed_id",
                    lamindb.base.fields.BigIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "doi",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                message="Must be a DOI (e.g., 10.1000/xyz123 or https://doi.org/10.1000/xyz123)",
                                regex="^(?:https?://(?:dx\\.)?doi\\.org/|doi:|DOI:)?10\\.\\d+/.*$",
                            )
                        ],
                    ),
                ),
                (
                    "preprint",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=False
                    ),
                ),
                (
                    "public",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                (
                    "journal",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "description",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "text",
                    lamindb.base.fields.TextField(blank=True, default=None, null=True),
                ),
                (
                    "published_at",
                    lamindb.base.fields.DateField(blank=True, default=None, null=True),
                ),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="references",
                        through="lamindb.ArtifactReference",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "authors",
                    models.ManyToManyField(
                        related_name="references", to="lamindb.person"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="references",
                        through="lamindb.CollectionReference",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                lamindb.models.CanCurate,
                models.Model,
                lamindb.models.ValidateFields,
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="references",
            field=models.ManyToManyField(
                related_name="projects", to="lamindb.reference"
            ),
        ),
        migrations.AddField(
            model_name="collectionreference",
            name="reference",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_collection",
                to="lamindb.reference",
            ),
        ),
        migrations.AddField(
            model_name="artifactreference",
            name="reference",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_artifact",
                to="lamindb.reference",
            ),
        ),
        migrations.CreateModel(
            name="TransformProject",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "project",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_transform",
                        to="lamindb.project",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "transform",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_project",
                        to="lamindb.transform",
                    ),
                ),
            ],
            options={
                "unique_together": {("transform", "project")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="project",
            name="transforms",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.TransformProject",
                to="lamindb.transform",
            ),
        ),
        migrations.CreateModel(
            name="TransformReference",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(auto_now_add=True, db_index=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "reference",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_transform",
                        to="lamindb.reference",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "transform",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_reference",
                        to="lamindb.transform",
                    ),
                ),
            ],
            options={
                "unique_together": {("transform", "reference")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="reference",
            name="transforms",
            field=models.ManyToManyField(
                related_name="references",
                through="lamindb.TransformReference",
                to="lamindb.transform",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="collectionproject",
            unique_together={("collection", "project")},
        ),
        migrations.AlterUniqueTogether(
            name="artifactproject",
            unique_together={("artifact", "project", "feature")},
        ),
        migrations.AlterUniqueTogether(
            name="collectionreference",
            unique_together={("collection", "reference")},
        ),
        migrations.AlterUniqueTogether(
            name="artifactreference",
            unique_together={("artifact", "reference", "feature")},
        ),
        migrations.RunPython(
            migrate_data,
        ),
    ]


if "ourprojects" in lamindb_setup.settings.instance.modules:
    Migration.dependencies += [
        ("ourprojects", "0003_alter_person_space_alter_project_space_and_more"),
    ]



"""Fields.

.. autosummary::
   :toctree: .

   CharField
   TextField
   ForeignKey
   BooleanField
   DateField
   DateTimeField
   BigIntegerField
   IntegerField
   OneToOneField
   FloatField
   DecimalField
   BinaryField
   JSONField
   EmailField
   TimeField
   SlugField
   URLField
   UUIDField
   PositiveIntegerField
   PositiveSmallIntegerField
   SmallIntegerField
   GenericIPAddressField
   DurationField
   CharField
   TextField
"""

from django.db import models


class CharField(models.CharField):
    """Custom `CharField` with default values for `blank`, `default`, and `max_length`.

    Django default values for `CharField` are `blank=False`, `default=""`, undefined `max_length`.
    """

    def __init__(self, max_length: int = 255, **kwargs):
        kwargs["max_length"] = max_length  # Set max_length in kwargs
        kwargs.setdefault("blank", True)
        kwargs.setdefault("default", None)
        super().__init__(**kwargs)  # Pass all arguments as kwargs


class TextField(models.TextField):
    """Custom `TextField` with default values for `blank` and `default`.

    Django default values for `TextField` are `blank=False`, `default=''`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("default", None)
        super().__init__(*args, **kwargs)


class ForeignKey(models.ForeignKey):
    """Custom `ForeignKey` with default values for `blank`.

    Django default value for `ForeignKey` `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


# fix doc string that otherwise errors
ForeignKey.get_extra_descriptor_filter.__doc__ = (
    ForeignKey.get_extra_descriptor_filter.__doc__.replace(
        ".filter(**kwargs)", "`.filter(**kwargs)`"
    )
)


class BooleanField(models.BooleanField):
    """Custom `BooleanField` with default values for `blank` and `default`.

    Django default values for `BooleanField` are `blank=False`, `default=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("default", None)
        super().__init__(*args, **kwargs)


class DateField(models.DateField):
    """Custom `DateField` with default values for `blank`.

    Django default values for `DateField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class DateTimeField(models.DateTimeField):
    """Custom `DateTimeField` with default values for `blank`.

    Django default values for `DateTimeField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class BigIntegerField(models.BigIntegerField):
    """Custom `BigIntegerField` with default values for `blank`.

    Django default values for `BigIntegerField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("default", None)
        super().__init__(*args, **kwargs)


class IntegerField(models.IntegerField):
    """Custom `IntegerField` with default values for `blank`.

    Django default values for `IntegerField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class OneToOneField(models.OneToOneField):
    """Custom `OneToOneField` with default values for `blank`.

    Django default values for `OneToOneField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class FloatField(models.FloatField):
    """Custom `FloatField` with default values for `blank`.

    Django default values for `FloatField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class DecimalField(models.DecimalField):
    """Custom `DecimalField` with default values for `blank`.

    Django default values for `DecimalField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class JSONField(models.JSONField):
    """Custom `JSONField` with default values for `blank`.

    Django default values for `JSONField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class DurationField(models.DurationField):
    """Custom `DurationField` with default values for `blank`.

    Django default values for `DurationField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class URLField(models.URLField):
    """Custom `URLField` with default values for `blank`.

    Django default values for `URLField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class EmailField(models.EmailField):
    """Custom `EmailField` with default values for `blank`.

    Django default values for `EmailField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class TimeField(models.TimeField):
    """Custom `TimeField` with default values for `blank`.

    Django default values for `TimeField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class SlugField(models.SlugField):
    """Custom `SlugField` with default values for `blank`.

    Django default values for `SlugField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class UUIDField(models.UUIDField):
    """Custom `UUIDField` with default values for `blank`.

    Django default values for `UUIDField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class PositiveIntegerField(models.PositiveIntegerField):
    """Custom `PositiveIntegerField` with default values for `blank`.

    Django default values for `PositiveIntegerField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class PositiveSmallIntegerField(models.PositiveSmallIntegerField):
    """Custom `PositiveSmallIntegerField` with default values for `blank`.

    Django default values for `PositiveSmallIntegerField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class SmallIntegerField(models.SmallIntegerField):
    """Custom `SmallIntegerField` with default values for `blank`.

    Django default values for `SmallIntegerField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class BinaryField(models.BinaryField):
    """Custom `BinaryField` with default values for `blank`.

    Django default values for `BinaryField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)


class GenericIPAddressField(models.GenericIPAddressField):
    """Custom `GenericIPAddressField` with default values for `blank`.

    Django default values for `GenericIPAddressField` are `blank=False`.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)



from __future__ import annotations

from typing import Any, Literal

import anndata as ad
import numpy as np
import pandas as pd


def small_dataset3_cellxgene(
    otype: Literal["DataFrame", "AnnData"] = "AnnData",
    *,
    with_obs_defaults: bool = False,
    with_var_typo: bool = False,
    with_obs_typo: bool = False,
    with_uns_organism: bool = False,
    with_uns_spatial: bool = False,
) -> tuple[pd.DataFrame, dict[str, Any]] | ad.AnnData:
    var_id = "invalid_ensembl_id" if with_var_typo else "ENSG00000000457"
    var_ids = [var_id, "ENSG00000000419", "ENSG00000139618"]
    lung_id = "UBERON:0002048XXX" if with_obs_typo else "UBERON:0002048"

    obs_data = {
        "disease_ontology_term_id": [
            "MONDO:0004975",
            "MONDO:0004980",
            "MONDO:0004980",
        ],
        "development_stage_ontology_term_id": ["unknown", "unknown", "unknown"],
        "sex_ontology_term_id": ["PATO:0000383", "PATO:0000384", "unknown"],
        "tissue_ontology_term_id": [lung_id, lung_id, "UBERON:0000948"],
        "cell_type": ["T cell", "B cell", "B cell"],
        "self_reported_ethnicity": ["South Asian", "South Asian", "South Asian"],
        "donor_id": ["-1", "1", "2"],
        "is_primary_data": [False, False, False],
        "suspension_type": ["cell", "cell", "cell"],
        "tissue_type": ["tissue", "tissue", "tissue"],
    }

    obs_df = pd.DataFrame(
        obs_data,
        index=["barcode1", "barcode2", "barcode3"],
    )

    var_df = pd.DataFrame(
        index=var_ids, data={"feature_is_filtered": [False, False, False]}
    )

    X = pd.DataFrame(
        {
            var_ids[0]: [2, 3, 3],
            var_ids[1]: [3, 4, 5],
            var_ids[2]: [4, 2, 3],
        },
        index=["barcode1", "barcode2", "barcode3"],
        dtype="float32",
    )

    obs_df["donor_id"] = obs_df["donor_id"].astype("category")

    if otype == "DataFrame":
        return pd.concat([X, obs_df], axis=1)
    else:
        adata = ad.AnnData(X=X, obs=obs_df, var=var_df)
        adata.uns["title"] = "CELLxGENE example"
        adata.obsm["X_pca"] = np.array(
            [[-1.2, 0.8], [0.5, -0.3], [0.7, -0.5]], dtype="float32"
        )
        # CELLxGENE requires the `.raw` slot to be set - https://github.com/chanzuckerberg/single-cell-curation/issues/1304
        adata.raw = adata.copy()
        adata.raw.var.drop(columns="feature_is_filtered", inplace=True)

        if with_obs_defaults:
            adata.obs["cell_type_ontology_term_id"] = [
                "CL:0000084",
                "CL:0000236",
                "CL:0000236",
            ]
            adata.obs["self_reported_ethnicity_ontology_term_id"] = "na"
            adata.obs["assay_ontology_term_id"] = "EFO:1001982"
            adata.obs["assay"] = "single-cell RNA sequencing"
        if with_uns_organism:
            adata.uns["organism_ontology_term_id"] = "NCBITaxon:9606"
            adata.uns["organism"] = "Homo sapiens"
        else:
            adata.obs["organism_ontology_term_id"] = "NCBITaxon:9606"
            obs_data["organism"] = ["Homo sapiens", "Homo sapiens", "Homo sapiens"]
        if with_uns_spatial:
            adata.uns["spatial"] = {
                "is_single": True,
                "library_123": {
                    "scalefactors": {
                        "spot_diameter_fullres": 165.0,
                        "tissue_hires_scalef": 0.5,
                    },
                    "images": {
                        "hires": np.random.default_rng().integers(
                            0, 255, (2000, 2000, 3), dtype=np.uint8
                        )
                    },
                },
            }

        return adata


def anndata_with_obs() -> ad.AnnData:
    """Create a mini anndata with cell_type, disease and tissue."""
    import anndata as ad
    import bionty.base as bionty_base

    celltypes = ["T cell", "hematopoietic stem cell", "hepatocyte", "my new cell type"]
    celltype_ids = ["CL:0000084", "CL:0000037", "CL:0000182", ""]
    diseases = [
        "chronic kidney disease",
        "liver lymphoma",
        "cardiac ventricle disorder",
        "Alzheimer disease",
    ]
    tissues = ["kidney", "liver", "heart", "brain"]
    df = pd.DataFrame()
    df["cell_type"] = celltypes * 10
    df["cell_type_id"] = celltype_ids * 10
    df["tissue"] = tissues * 10
    df["disease"] = diseases * 10
    df.index = "obs" + df.index.astype(str)

    adata = ad.AnnData(X=np.zeros(shape=(40, 100), dtype=np.float32), obs=df)
    bionty_genes = bionty_base.Gene()
    # backwards compatible
    adata.var.index = (
        (
            bionty_genes.to_dataframe()
            if hasattr(bionty_genes, "to_dataframe")
            else bionty_genes.df()
        )
        .head(100)["ensembl_gene_id"]
        .values
    )

    return adata



import lamindb as ln
import pandas as pd
import pytest
from lamindb.models._is_versioned import bump_version, set_version


@pytest.fixture(scope="module")
def df1():
    return pd.DataFrame({"feat1": [1, 2]})


@pytest.fixture(scope="module")
def df2():
    return pd.DataFrame({"feat1": [2, 3]})


def test_set_version():
    # all remaining lines are covered in notebooks
    with pytest.raises(ValueError):
        set_version(None, "weird-version")
    assert set_version(None, "1.2") == "2"
    assert set_version(None, "0") == "1"
    assert set_version(None, "1") == "2"
    assert set_version("1.2.3", "0") == "1.2.3"
    assert set_version("1.2.3") == "1.2.3"


def test_bump_version():
    current_version_major_only = "2"
    current_version_major_minor = "2.1"
    weird_version = "weird-version"
    with pytest.raises(ValueError):
        bump_version(weird_version)
    assert bump_version(weird_version, behavior="ignore") == "?"
    assert bump_version(current_version_major_only, bump_type="major") == "3"
    assert bump_version(current_version_major_only, bump_type="minor") == "2.1"
    assert bump_version(current_version_major_minor, bump_type="major") == "3"
    assert bump_version(current_version_major_minor, bump_type="minor") == "2.2"


def test_add_to_version_family(df1, df2):
    artifact1 = ln.Artifact.from_dataframe(df1, description="test1")
    artifact1.save()
    artifact2 = ln.Artifact.from_dataframe(df2, description="test2")
    artifact2.save()
    assert (
        artifact1.uid[: artifact1._len_stem_uid]
        != artifact2.uid[: artifact2._len_stem_uid]
    )
    artifact2._add_to_version_family(artifact1)
    assert (
        artifact1.uid[: artifact1._len_stem_uid]
        == artifact2.uid[: artifact2._len_stem_uid]
    )
    assert (
        artifact1.path.name[: artifact1._len_stem_uid]
        == artifact2.path.name[: artifact2._len_stem_uid]
    )
    artifact1.delete(permanent=True)
    artifact2.delete(permanent=True)


def test_transform_versioning_based_on_key():
    transform1 = ln.Transform(
        key="test-pipeline",
        version="1.0",
        source_code="1",
        type="pipeline",
    ).save()
    assert transform1.is_latest

    with pytest.raises(ValueError) as e:
        transform2 = ln.Transform(
            key="test-pipeline",
            version="1.0",
            source_code="2",
            type="pipeline",
        ).save()
    assert (
        e.exconly()
        == "ValueError: Please change the version tag or leave it `None`, '1.0' is already taken"
    )

    transform2 = ln.Transform(
        key="test-pipeline",
        # do not pass the version tag, which corresponds to: version=None
        source_code="2",
        type="pipeline",
    ).save()

    assert transform2.version is None
    assert transform2.is_latest
    assert transform2.hash != transform1.hash
    assert not ln.Transform.get(key="test-pipeline", version="1.0").is_latest

    transform3 = ln.Transform(
        key="test-pipeline",
        version="abcd",  # mimic commit hash
        source_code="3",
        type="pipeline",
    ).save()

    assert transform3.version == "abcd"
    assert transform3.is_latest
    assert transform3.hash != transform2.hash
    assert not ln.Transform.get(key="test-pipeline", source_code="2").is_latest


def test_transform_versioning_based_on_revises():
    # build one version family
    transform_v1 = ln.Transform(description="Introduction").save()
    assert transform_v1.is_latest
    assert transform_v1.version is None

    # pass the latest version
    transform_v2 = ln.Transform(
        description="Introduction v2", revises=transform_v1, version="2"
    ).save()
    assert not transform_v1.is_latest
    assert transform_v2.is_latest
    assert transform_v2.uid.endswith("0001")

    # consciously *not* pass the latest version to revises but the previous
    # it automatically retrieves the latest version
    transform_v3 = ln.Transform(description="Introduction", revises=transform_v1).save()
    assert transform_v3.uid.endswith("0002")
    assert not ln.Transform.objects.get(
        description="Introduction v2", version="2"
    ).is_latest
    assert transform_v3.is_latest
    transform_v4 = ln.Transform(description="Introduction").save()
    assert transform_v4.is_latest

    # add another transform with the same description that's not part of this family
    # but will also be a hit for the query
    assert len(ln.Transform.filter(description="Introduction")) == 3
    assert len(ln.Transform.filter(description="Introduction").latest_version()) == 2
    transform_v4.delete(permanent=True)
    assert ln.Transform.get(description="Introduction") == transform_v3
    assert (
        ln.Transform.filter(description="Introduction").latest_version().one()
        == transform_v3
    )

    # test get
    assert ln.Transform.get(transform_v3.uid) == transform_v3
    assert ln.Transform.get(transform_v3.id) == transform_v3
    assert ln.Transform.get(transform_v3.uid[:-4]) == transform_v3

    # test empty QuerySet
    assert (
        ln.Transform.filter(description="IntroductionNotExists")
        .latest_version()
        .one_or_none()
        is None
    )

    # test soft delete
    transform_v3.delete()
    assert transform_v2.is_latest

    # test hard delete
    transform_v2.delete(permanent=True)
    assert (
        transform_v1_retrieved := ln.Transform.get(transform_v3.uid[:-4])
    ) == transform_v1
    assert transform_v1_retrieved.is_latest

    # test soft delete on the last existing version does not change is_latest
    transform_v1_retrieved.delete()
    assert (
        transform_v1_retrieved := ln.Transform.get(transform_v1.uid)
    ) == transform_v1
    assert transform_v1_retrieved.is_latest

    # fully delete
    transform_v1.delete(permanent=True)

    # last object that exists is in the trash
    assert ln.Transform.get(transform_v3.uid[:-4]) == transform_v3
    assert transform_v3.branch_id == -1
    transform_v3.delete(permanent=True)


def test_path_rename():
    # this is related to renames inside _add_to_version_family
    with open("test_new_path.txt", "w") as f:
        f.write("test_new_path")
    old_path = ln.UPath("s3://lamindata/.lamindb/test_new_path.txt")
    old_path.upload_from("./test_new_path.txt")
    assert old_path.exists()
    new_path = old_path.rename(old_path.with_name("test_new_path2.txt"))
    assert new_path.exists()
    assert new_path.as_posix() == "s3://lamindata/.lamindb/test_new_path2.txt"
    assert not old_path.exists()
    new_path.unlink()
    ln.UPath("./test_new_path.txt").unlink()



import lamindb as ln
import pytest


@pytest.mark.parametrize("key", [None, "my_new_folder"])
def test_folder_like_artifact(get_test_filepaths, key):
    # get variables from fixture
    is_in_registered_storage = get_test_filepaths[0]
    test_dirpath = get_test_filepaths[2]
    hash_test_dir = get_test_filepaths[5]
    if key is None and not is_in_registered_storage:
        with pytest.raises(ValueError) as error:
            ln.Artifact(test_dirpath, key=key)
        assert error.exconly().startswith(
            "ValueError: Pass one of key, run or description as a parameter"
        )
        return None
    artifact1 = ln.Artifact(test_dirpath, key=key)
    if key is not None and is_in_registered_storage:
        assert artifact1._real_key is not None
    else:
        assert artifact1._real_key is None
    assert artifact1.n_files == 3
    assert artifact1.hash == hash_test_dir
    assert artifact1._state.adding
    assert artifact1.description is None
    assert artifact1.path.exists()
    artifact1.save()

    # run tests on re-creating the Artifact
    artifact2 = ln.Artifact(test_dirpath, key=key, description="something")
    assert not artifact2._state.adding
    assert artifact1.id == artifact2.id
    assert artifact1.uid == artifact2.uid
    assert artifact1.storage == artifact2.storage
    assert artifact2.path.exists()
    assert artifact2.description == "something"

    # now put another file in the test directory

    # create a first file
    test_filepath_added = test_dirpath / "my_file_added.txt"
    test_filepath_added.write_text("2")
    artifact3 = ln.Artifact(test_dirpath, key=key, revises=artifact1)
    assert artifact3.n_files == 4
    assert artifact3.hash != hash_test_dir
    assert artifact3._state.adding
    assert artifact3.description is None
    assert artifact3.path.exists()
    artifact3.save()

    # the state of artifact1 is lost, because artifact3 is stored at the same path
    assert artifact3.overwrite_versions
    assert artifact1.overwrite_versions
    assert artifact3.path == artifact1.path
    test_filepath_added.unlink()

    # delete the artifact
    artifact2.delete(permanent=True, storage=False)
    artifact3.delete(permanent=True, storage=False)


def test_overwrite_versions_false(get_test_filepaths):
    # get variables from fixture
    is_in_registered_storage = get_test_filepaths[0]
    test_dirpath = get_test_filepaths[2]
    hash_test_dir = get_test_filepaths[5]
    if is_in_registered_storage:
        return
    artifact1 = ln.Artifact(
        test_dirpath, key="my_folder", overwrite_versions=False
    ).save()
    assert artifact1.hash == hash_test_dir
    # skip artifact2 because we already test this above
    # create a first file
    test_filepath_added = test_dirpath / "my_file_added.txt"
    test_filepath_added.write_text("2")
    artifact3 = ln.Artifact(test_dirpath, key="my_folder", overwrite_versions=False)
    assert artifact3.hash != hash_test_dir
    artifact3.save()
    # the state of artifact1 is lost, because artifact3 is stored at the same path
    assert not artifact3.overwrite_versions
    assert not artifact1.overwrite_versions
    assert artifact3.path != artifact1.path
    test_filepath_added.unlink()
    artifact1.delete(permanent=True, storage=False)
    artifact3.delete(permanent=True, storage=False)



from datetime import datetime
from typing import Any, Callable, Iterable

import pandas as pd


def is_list_of_type(value: Any, expected_type: Any) -> bool:
    """Helper function to check if a value is either of expected_type or a list of that type, or a mix of both in a nested structure."""
    if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
        # handle nested lists recursively
        return all(is_list_of_type(item, expected_type) for item in value)
    return isinstance(value, expected_type)


def check_dtype(expected_type: Any) -> Callable:
    """Creates a check function for Pandera that validates a column's dtype.

    Supports both standard dtype checking and mixed list/single values for the same type.
    For example, a column with expected_type 'float' would also accept a mix of float values and lists of floats.

    Args:
        expected_type: String identifier for the expected type ('int', 'float', 'num', 'str')

    Returns:
        A function that checks if a series has the expected dtype or contains mixed types
    """

    def check_function(series):
        # first check if the series is entirely of the expected dtype (fast path)
        if expected_type == "int" and pd.api.types.is_integer_dtype(series.dtype):
            return True
        elif expected_type == "float" and pd.api.types.is_float_dtype(series.dtype):
            return True
        elif expected_type == "num" and pd.api.types.is_numeric_dtype(series.dtype):
            return True
        elif expected_type == "str" and pd.api.types.is_string_dtype(series.dtype):
            return True
        elif expected_type == "path" and pd.api.types.is_string_dtype(series.dtype):
            return True

        # if we're here, it might be a mixed column with object dtype
        # need to check each value individually
        if series.dtype == "object" and expected_type.startswith("list"):
            expected_type_member = expected_type.replace("list[", "").removesuffix("]")
            if expected_type_member == "int":
                return series.apply(lambda x: is_list_of_type(x, int)).all()
            elif expected_type_member == "float":
                return series.apply(lambda x: is_list_of_type(x, float)).all()
            elif expected_type_member == "num":
                # for numeric, accept either int or float
                return series.apply(lambda x: is_list_of_type(x, (int, float))).all()
            elif (
                expected_type_member == "str"
                or expected_type_member == "path"
                or expected_type_member.startswith("cat[")
            ):
                return series.apply(lambda x: is_list_of_type(x, str)).all()

        # if we get here, the validation failed
        return False

    return check_function


def is_valid_datetime_str(date_string: str) -> bool | str:
    try:
        dt = datetime.fromisoformat(date_string)
        return dt.isoformat()
    except ValueError:
        return False


def is_iterable_of_sqlrecord(value: Any):
    from lamindb.models import SQLRecord

    return isinstance(value, Iterable) and isinstance(next(iter(value)), SQLRecord)



# Generated by Django 5.1.12 on 2025-10-02 15:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0130_branch_space_alter_artifactblock_artifact_and_more"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="record",
            constraint=models.UniqueConstraint(
                fields=("name", "type", "space"), name="unique_name_type_space"
            ),
        ),
    ]



from pathlib import Path
from typing import Callable, Generator

import lamindb as ln
import lightning as pl
import pytest
import torch
from lamindb.integrations import lightning as ll
from torch import nn
from torch.utils.data import DataLoader, TensorDataset


@pytest.fixture(autouse=True)
def cleanup_checkpoints() -> Generator[None, None, None]:
    yield
    for path in Path().glob("test_checkpoint*.ckpt"):
        path.unlink(missing_ok=True)


@pytest.fixture
def simple_model() -> pl.LightningModule:
    class SimpleModel(pl.LightningModule):
        def __init__(self):
            super().__init__()
            self.layer = nn.Linear(10, 1)

        def forward(self, x):
            return self.layer(x)

        def training_step(self, batch, batch_idx):
            x, y = batch
            loss = nn.functional.mse_loss(self(x), y)
            self.log("train_loss", loss)
            return loss

        def configure_optimizers(self):
            return torch.optim.Adam(self.parameters())

    model = SimpleModel()
    return model


@pytest.fixture
def torch_train_data_dataloader() -> DataLoader:
    train_data = DataLoader(
        TensorDataset(torch.randn(100, 10), torch.randn(100, 1)), batch_size=10
    )
    return train_data


def test_callback_basic(
    cleanup_checkpoints: Callable,
    torch_train_data_dataloader: DataLoader,
    simple_model: pl.LightningModule,
):
    """Callback should create artifacts for each training epoch."""
    artifact_key = "test/model.ckpt"

    callback = ll.Callback(path="test_checkpoint.ckpt", key=artifact_key)

    trainer = pl.Trainer(
        max_epochs=2, callbacks=[callback], enable_checkpointing=False, logger=False
    )
    trainer.fit(simple_model, torch_train_data_dataloader)

    artifacts = ln.Artifact.filter(key=artifact_key)
    assert len(artifacts) == 2

    for af in artifacts:
        assert af.kind == "model"
        af.delete(permanent=True)


def test_callback_with_features(
    cleanup_checkpoints: Callable,
    torch_train_data_dataloader: DataLoader,
    simple_model: pl.LightningModule,
):
    """Callback should annotate artifacts with feature values."""
    train_loss = ln.Feature(name="train_loss", dtype="float").save()
    custom_param = ln.Feature(name="custom_param", dtype="str").save()

    artifact_key = "test/model_features.ckpt"

    callback = ll.Callback(
        path="test_checkpoint_features.ckpt",
        key=artifact_key,
        features={"train_loss": None, "custom_param": "test_value"},
    )

    trainer = pl.Trainer(
        max_epochs=2, callbacks=[callback], enable_checkpointing=False, logger=False
    )
    trainer.fit(simple_model, torch_train_data_dataloader)

    artifacts = ln.Artifact.filter(key=artifact_key)
    assert len(artifacts) == 2

    for af in artifacts:
        values = af.features.get_values()
        assert "train_loss" in values
        assert values["custom_param"] == "test_value"
        af.delete(permanent=True)

    train_loss.delete(permanent=True)
    custom_param.delete(permanent=True)


def test_callback_missing_features(
    cleanup_checkpoints: Callable,
    torch_train_data_dataloader: DataLoader,
    simple_model: pl.LightningModule,
):
    """Callback should raise an error when specified features do not exist."""
    artifact_key = "test/model_missing.ckpt"

    callback = ll.Callback(
        path="test_checkpoint_missing.ckpt",
        key=artifact_key,
        features={"nonexistent_feature": None},
    )

    trainer = pl.Trainer(
        max_epochs=1, callbacks=[callback], enable_checkpointing=False, logger=False
    )

    with pytest.raises(ValueError) as e:
        trainer.fit(simple_model, torch_train_data_dataloader)
    assert "Feature nonexistent_feature missing" in str(e.value)



# Generated by Django 5.2 on 2025-05-22 15:21

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.uids
import lamindb.base.users
import lamindb.models.can_curate
import lamindb.models.run
import lamindb.models.sqlrecord


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0097_remove_schemaparam_param_remove_paramvalue_param_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="feature",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instances",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instances",
                to="lamindb.project",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instances",
                to="lamindb.reference",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instances",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="type",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instances",
                to="lamindb.ulabel",
            ),
        ),
        migrations.CreateModel(
            name="Record",
            fields=[
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                (
                    "_aux",
                    lamindb.base.fields.JSONField(
                        blank=True, db_default=None, default=None, null=True
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.uids.base62_16,
                        editable=False,
                        max_length=16,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        null=True,
                    ),
                ),
                (
                    "is_type",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=False, null=True
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
                (
                    "type",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="instances",
                        to="lamindb.record",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(lamindb.models.can_curate.CanCurate, models.Model),
        ),
        migrations.CreateModel(
            name="RecordArtifact",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordartifact",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_artifact",
                        to="lamindb.record",
                    ),
                ),
                (
                    "value",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.artifact",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature", "value")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="record",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="records",
                through="lamindb.RecordArtifact",
                to="lamindb.artifact",
            ),
        ),
        migrations.CreateModel(
            name="RecordProject",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordproject",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_project",
                        to="lamindb.record",
                    ),
                ),
                (
                    "value",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.project",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="project",
            name="records",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.RecordProject",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordRecord",
            fields=[
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                (
                    "_aux",
                    lamindb.base.fields.JSONField(
                        blank=True, db_default=None, default=None, null=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordrecord",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_record",
                        to="lamindb.record",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
                (
                    "value",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.record",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="record",
            name="components",
            field=models.ManyToManyField(
                related_name="composites",
                through="lamindb.RecordRecord",
                to="lamindb.record",
            ),
        ),
        migrations.CreateModel(
            name="RecordRun",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordrun",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_run",
                        to="lamindb.record",
                    ),
                ),
                (
                    "value",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="record",
            name="runs",
            field=models.ManyToManyField(
                related_name="records", through="lamindb.RecordRun", to="lamindb.run"
            ),
        ),
        migrations.CreateModel(
            name="RecordULabel",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordulabel",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_ulabel",
                        to="lamindb.record",
                    ),
                ),
                (
                    "value",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.ulabel",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="record",
            name="ulabels",
            field=models.ManyToManyField(
                related_name="_records",
                through="lamindb.RecordULabel",
                to="lamindb.ulabel",
            ),
        ),
        migrations.CreateModel(
            name="Sheet",
            fields=[
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                (
                    "_aux",
                    lamindb.base.fields.JSONField(
                        blank=True, db_default=None, default=None, null=True
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.uids.base62_12,
                        editable=False,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, db_index=True, default=None, max_length=255
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "schema",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sheets",
                        to="lamindb.schema",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="record",
            name="sheet",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="records",
                to="lamindb.sheet",
            ),
        ),
        migrations.CreateModel(
            name="SheetProject",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "project",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_sheet",
                        to="lamindb.project",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "sheet",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_project",
                        to="lamindb.sheet",
                    ),
                ),
            ],
            options={
                "unique_together": {("sheet", "project")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="project",
            name="sheets",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.SheetProject",
                to="lamindb.sheet",
            ),
        ),
        migrations.CreateModel(
            name="RecordJson",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "value",
                    lamindb.base.fields.JSONField(
                        blank=True, db_default=None, default=None
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_recordjson",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values_json",
                        to="lamindb.record",
                    ),
                ),
            ],
            options={
                "unique_together": {("record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
    ]



# Generated by Django 5.2 on 2025-02-06 07:10

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0085_alter_feature_is_type_alter_flextable_is_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transform",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.CreateModel(
            name="Migration",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "app",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255
                    ),
                ),
                ("applied", lamindb.base.fields.DateTimeField(blank=True)),
            ],
            options={
                "db_table": "django_migrations",
                "managed": False,
            },
        ),
        migrations.AlterField(
            model_name="param",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=64, null=True
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=255, null=True
            ),
        ),
    ]



import bionty as bt
import lamindb as ln
import pandas as pd
import pytest
from lamindb.core.exceptions import ValidationError


@pytest.fixture(scope="module")
def lists_df():
    return pd.DataFrame(
        {
            "sample_id": [["sample1", "sample2"], ["sample2"], ["sample3"]],
            "dose": [[1.2, 2.3], [1.2], [2.3]],
            "cell_type": [["B cell", "T cell"], ["B cell"], ["T cell"]],
            "tissue": [["blood", "pulmo"], ["blood"], ["lung"]],
        }
    )


@pytest.fixture(scope="module")
def nested_cat_df():
    return pd.DataFrame(
        {
            "biosample_id": ["S1", "S2", "S3", "S4", "S5", "S6"],
            "biosample_name": [
                "sample1",
                "sample2",
                "sample3",
                "sample4",
                "sample5",
                "sample6",
            ],
        }
    )


@pytest.fixture(scope="module")
def lists_schema():
    schema = ln.Schema(
        name="lists schema cat",
        features=[
            ln.Feature(name="sample_id", dtype=list[str]).save(),
            ln.Feature(name="dose", dtype=list[float]).save(),
            ln.Feature(name="cell_type", dtype=list[str]).save(),
            ln.Feature(name="tissue", dtype=list[bt.Tissue]).save(),
        ],
    ).save()

    yield schema

    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)
    bt.Tissue.filter().delete(permanent=True)


@pytest.fixture(scope="module")
def nested_cat_schema():
    schema = ln.Schema(
        name="schema_with_nested_cat",
        features=[
            ln.Feature(name="biosample_id", dtype=str).save(),
            ln.Feature(name="biosample_name", dtype="cat[Record[LabA[Sample]]]").save(),
        ],
        coerce_dtype=True,
    ).save()

    yield schema

    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)
    ln.Record.filter().update(type=None)
    ln.Record.filter().delete(permanent=True)


def test_curator_df_multivalue(lists_df, lists_schema):
    curator = ln.curators.DataFrameCurator(lists_df, lists_schema)
    with pytest.raises(ValidationError):
        curator.validate()
    assert curator.cat._cat_vectors.keys() == {"columns", "tissue"}
    assert curator.cat._cat_vectors["tissue"]._validated == ["blood", "lung"]
    assert curator.cat._cat_vectors["tissue"]._non_validated == ["pulmo"]
    assert curator.cat._cat_vectors["tissue"]._synonyms == {"pulmo": "lung"}

    curator.cat.standardize("tissue")
    assert curator.cat._cat_vectors["tissue"]._non_validated == []
    assert lists_df["tissue"].tolist() == [["blood", "lung"], ["blood"], ["lung"]]

    assert curator.validate() is None


def test_curators_df_nested_cat(nested_cat_df, nested_cat_schema):
    # note that there are two type records both called "Sample" but having different parent types
    # first we create LabA->Sample
    LabA = ln.Record(name="LabA", is_type=True).save()
    Sample = ln.Record(name="Sample", is_type=True, type=LabA).save()
    for name in ["sample1", "sample2", "sample3", "sample4"]:
        ln.Record(name=name, type=Sample).save()

    # then we create LabB->Sample
    LabB = ln.Record(name="LabB", is_type=True).save()
    Sample = ln.Record(name="Sample", is_type=True, type=LabB).save()
    for name in ["sample5", "sample6"]:
        ln.Record(name=name, type=Sample).save()

    # "sample5" is not part of LabA, so it should not be validated
    with pytest.raises(ValidationError):
        curator = ln.curators.DataFrameCurator(nested_cat_df, nested_cat_schema)
        curator.validate()

    assert len(curator.cat._cat_vectors["biosample_name"]._validated) == 4
    assert len(curator.cat._cat_vectors["biosample_name"]._non_validated) == 2



# Generated by Django 5.1.12 on 2025-10-01 16:44

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0128_artifact__real_key"),
    ]

    operations = [
        migrations.CreateModel(
            name="ArtifactBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "artifact",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BranchBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "branch",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.branch",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CollectionBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FeatureBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.feature",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ProjectBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.project",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RecordBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "record",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.record",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RootBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=255)),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RunBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.run",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SchemaBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "schema",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.schema",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SpaceBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "space",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.space",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TransformBlock",
            fields=[
                (
                    "version",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "is_latest",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=True
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=20, unique=True
                    ),
                ),
                ("content", models.TextField()),
                ("hash", models.CharField(db_index=True, max_length=22, null=True)),
                (
                    "type",
                    models.CharField(
                        db_default="mdpage",
                        db_index=True,
                        default="mdpage",
                        max_length=22,
                    ),
                ),
                (
                    "vertical_pos",
                    models.SmallIntegerField(db_default=0, db_index=True, default=0),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("line_number", models.IntegerField(null=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "transform",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blocks",
                        to="lamindb.transform",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 5.2 on 2025-01-05 11:58

from pathlib import Path

import lamindb_setup as ln_setup
import psycopg2
from django.db import migrations


def get_artifact_path_psycopg2(artifact_id):
    """Get artifact path using psycopg2."""
    query = """
        SELECT
            s.root || '/.lamindb/' || a.uid || a.suffix AS full_path
        FROM
            lamindb_artifact a
            JOIN lamindb_storage s ON a.storage_id = s.id
        WHERE
            a.id = %s
    """

    with psycopg2.connect(ln_setup.settings.instance.db) as conn:
        with conn.cursor() as cur:
            cur.execute(query, (artifact_id,))
            return cur.fetchone()[0]


def transfer_source_code(apps, schema_editor):
    from lamindb._finish import notebook_to_script

    Transform = apps.get_model("lamindb", "Transform")
    transforms = Transform.objects.filter(
        _source_code_artifact__isnull=False,
    ).select_related("_source_code_artifact")

    for transform in transforms:
        print(f"migrating source code of transform {transform}")
        artifact = transform._source_code_artifact
        print("artifact", artifact.uid)

        path_str = get_artifact_path_psycopg2(artifact.id)
        print(ln_setup.settings.storage.root_as_str)
        print(path_str)
        if path_str.startswith(ln_setup.settings.storage.root_as_str):
            path = (
                ln_setup.settings.storage.root
                / f".lamindb/{artifact.uid}{artifact.suffix}"
            )
        else:
            path = ln_setup.core.upath.UPath(path_str)
        if path.exists():
            if path_str.startswith("s3://"):
                local_path = Path(f"temp{path.suffix}")
                path.download_to(local_path)
            else:
                local_path = path

            if artifact.suffix == ".ipynb":
                transform.source_code = notebook_to_script(transform, local_path)
            else:
                transform.source_code = local_path.read_text()
            transform.hash = artifact.hash
            path.unlink()
        else:
            print(f"path did not exist: {path_str}")
        transform._source_code_artifact = None
        transform.save()
        artifact.delete()


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0069_squashed"),
    ]

    operations = [
        migrations.RunPython(transfer_source_code),
    ]



# Generated by Django 5.2 on 2025-07-06 23:09

from django.db import migrations

import lamindb.base.fields


def extenddefault_values(apps, schema_editor):
    """Lowercase default values for Space and Branch models."""
    Space = apps.get_model("lamindb", "Space")
    Branch = apps.get_model("lamindb", "Branch")

    space = Space.objects.get(uid="a")
    space.uid = 12 * "a"
    space.save()

    trash_branch = Branch.objects.get(uid="t")
    trash_branch.uid = 12 * "t"
    trash_branch.save()

    archive_branch = Branch.objects.get(uid="a")
    archive_branch.uid = 12 * "a"
    archive_branch.save()

    main_branch = Branch.objects.get(uid="m")
    main_branch.uid = 12 * "m"
    main_branch.save()


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0114_alter_run__status_code"),
    ]

    operations = [
        migrations.AlterField(
            model_name="space",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_default="aaaaaaaaaaaa",
                db_index=True,
                default="aaaaaaaaaaaaa",
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.RunPython(
            extenddefault_values,
        ),
    ]
